@@ -3,7 +3,7 @@
 #
 # SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
-# General Public License as public by the Free Software Foundation; version 2.0
+# General Public License as published by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
 # modify it under the terms of either of these two licenses.
 
@@ -18,9 +18,15 @@ https://git-scm.com/docs/gitformat-commit-graph
 
 import os
 import struct
-from typing import BinaryIO, Optional, Union
+from collections.abc import Iterator
+from typing import TYPE_CHECKING, BinaryIO, Optional, Union
 
-from .objects import ObjectID, hex_to_sha, sha_to_hex
+from .file import _GitFile
+
+if TYPE_CHECKING:
+    from .object_store import BaseObjectStore
+
+from .objects import Commit, ObjectID, hex_to_sha, sha_to_hex
 
 # File format constants
 COMMIT_GRAPH_SIGNATURE = b"CGPH"
@@ -62,6 +68,15 @@ class CommitGraphEntry:
         generation: int,
         commit_time: int,
     ) -> None:
+        """Initialize CommitGraphEntry.
+
+        Args:
+          commit_id: The commit object ID
+          tree_id: The tree object ID
+          parents: List of parent commit IDs
+          generation: Generation number
+          commit_time: Commit timestamp
+        """
         self.commit_id = commit_id
         self.tree_id = tree_id
         self.parents = parents
@@ -69,6 +84,7 @@ class CommitGraphEntry:
         self.commit_time = commit_time
 
     def __repr__(self) -> str:
+        """Return string representation of CommitGraphEntry."""
         return (
             f"CommitGraphEntry(commit_id={self.commit_id!r}, "
             f"tree_id={self.tree_id!r}, parents={self.parents!r}, "
@@ -80,10 +96,17 @@ class CommitGraphChunk:
     """Represents a chunk in the commit graph file."""
 
     def __init__(self, chunk_id: bytes, data: bytes) -> None:
+        """Initialize CommitGraphChunk.
+
+        Args:
+          chunk_id: Chunk identifier
+          data: Chunk data
+        """
         self.chunk_id = chunk_id
         self.data = data
 
     def __repr__(self) -> str:
+        """Return string representation of CommitGraphChunk."""
         return f"CommitGraphChunk(chunk_id={self.chunk_id!r}, size={len(self.data)})"
 
 
@@ -91,6 +114,11 @@ class CommitGraph:
     """Git commit graph file reader/writer."""
 
     def __init__(self, hash_version: int = HASH_VERSION_SHA1) -> None:
+        """Initialize CommitGraph.
+
+        Args:
+          hash_version: Hash version to use (SHA1 or SHA256)
+        """
         self.hash_version = hash_version
         self.chunks: dict[bytes, CommitGraphChunk] = {}
         self.entries: list[CommitGraphEntry] = []
@@ -265,7 +293,7 @@ class CommitGraph:
         entry = self.get_entry_by_oid(oid)
         return entry.parents if entry else None
 
-    def write_to_file(self, f: BinaryIO) -> None:
+    def write_to_file(self, f: Union[BinaryIO, _GitFile]) -> None:
         """Write commit graph to file."""
         if not self.entries:
             raise ValueError("Cannot write empty commit graph")
@@ -358,7 +386,7 @@ class CommitGraph:
         """Return number of commits in the graph."""
         return len(self.entries)
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator["CommitGraphEntry"]:
         """Iterate over commit graph entries."""
         return iter(self.entries)
 
@@ -396,7 +424,9 @@ def find_commit_graph_file(git_dir: Union[str, bytes]) -> Optional[bytes]:
     return None
 
 
-def generate_commit_graph(object_store, commit_ids: list[ObjectID]) -> CommitGraph:
+def generate_commit_graph(
+    object_store: "BaseObjectStore", commit_ids: list[ObjectID]
+) -> CommitGraph:
     """Generate a commit graph from a set of commits.
 
     Args:
@@ -426,12 +456,13 @@ def generate_commit_graph(object_store, commit_ids: list[ObjectID]) -> CommitGra
             normalized_commit_ids.append(commit_id)
 
     # Build a map of all commits and their metadata
-    commit_map = {}
+    commit_map: dict[bytes, Commit] = {}
     for commit_id in normalized_commit_ids:
         try:
             commit_obj = object_store[commit_id]
             if commit_obj.type_name != b"commit":
                 continue
+            assert isinstance(commit_obj, Commit)
             commit_map[commit_id] = commit_obj
         except KeyError:
             # Commit not found, skip
@@ -440,7 +471,7 @@ def generate_commit_graph(object_store, commit_ids: list[ObjectID]) -> CommitGra
     # Calculate generation numbers using topological sort
     generation_map: dict[bytes, int] = {}
 
-    def calculate_generation(commit_id):
+    def calculate_generation(commit_id: ObjectID) -> int:
         if commit_id in generation_map:
             return generation_map[commit_id]
 
@@ -507,7 +538,9 @@ def generate_commit_graph(object_store, commit_ids: list[ObjectID]) -> CommitGra
 
 
 def write_commit_graph(
-    git_dir: Union[str, bytes], object_store, commit_ids: list[ObjectID]
+    git_dir: Union[str, bytes],
+    object_store: "BaseObjectStore",
+    commit_ids: list[ObjectID],
 ) -> None:
     """Write a commit graph file for the given commits.
 
@@ -538,7 +571,7 @@ def write_commit_graph(
 
 
 def get_reachable_commits(
-    object_store, start_commits: list[ObjectID]
+    object_store: "BaseObjectStore", start_commits: list[ObjectID]
 ) -> list[ObjectID]:
     """Get all commits reachable from the given starting commits.
 
@@ -578,7 +611,7 @@ def get_reachable_commits(
 
         try:
             commit_obj = object_store[commit_id]
-            if commit_obj.type_name != b"commit":
+            if not isinstance(commit_obj, Commit):
                 continue
 
             # Add to reachable list (commit_id is already hex ObjectID)
