@@ -1,8 +1,9 @@
 # requests_vendor.py -- requests implementation of the AbstractHttpGitClient interface
 # Copyright (C) 2022 Eden Shalit <epopcop@gmail.com>
 #
+# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
-# General Public License as public by the Free Software Foundation; version 2.0
+# General Public License as published by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
 # modify it under the terms of either of these two licenses.
 #
@@ -29,7 +30,12 @@ the dulwich.client.HttpGitClient attribute:
 
 This implementation is experimental and does not have any tests.
 """
+
 from io import BytesIO
+from typing import TYPE_CHECKING, Any, Callable, Optional
+
+if TYPE_CHECKING:
+    from ..config import ConfigFile
 
 from requests import Session
 
@@ -43,28 +49,47 @@ from ..errors import GitProtocolError, NotGitRepository
 
 
 class RequestsHttpGitClient(AbstractHttpGitClient):
+    """HTTP Git client using the requests library."""
+
     def __init__(
-            self,
-            base_url,
-            dumb=None,
-            config=None,
-            username=None,
-            password=None,
-            **kwargs
+        self,
+        base_url: str,
+        dumb: Optional[bool] = None,
+        config: Optional["ConfigFile"] = None,
+        username: Optional[str] = None,
+        password: Optional[str] = None,
+        **kwargs: object,
     ) -> None:
+        """Initialize RequestsHttpGitClient.
+
+        Args:
+          base_url: Base URL of the Git repository
+          dumb: Whether to use dumb HTTP transport
+          config: Git configuration file
+          username: Username for authentication
+          password: Password for authentication
+          **kwargs: Additional keyword arguments
+        """
         self._username = username
         self._password = password
 
         self.session = get_session(config)
 
         if username is not None:
-            self.session.auth = (username, password)
+            self.session.auth = (username, password)  # type: ignore[assignment]
 
         super().__init__(
-            base_url=base_url, dumb=dumb, **kwargs)
+            base_url=base_url, dumb=bool(dumb) if dumb is not None else False, **kwargs
+        )
 
-    def _http_request(self, url, headers=None, data=None, allow_compression=False):
-        req_headers = self.session.headers.copy()
+    def _http_request(
+        self,
+        url: str,
+        headers: Optional[dict[str, str]] = None,
+        data: Optional[bytes] = None,
+        allow_compression: bool = False,
+    ) -> tuple[Any, Callable[[int], bytes]]:
+        req_headers = self.session.headers.copy()  # type: ignore[attr-defined]
         if headers is not None:
             req_headers.update(headers)
 
@@ -79,45 +104,54 @@ class RequestsHttpGitClient(AbstractHttpGitClient):
             resp = self.session.get(url, headers=req_headers)
 
         if resp.status_code == 404:
-            raise NotGitRepository()
+            raise NotGitRepository
         if resp.status_code == 401:
             raise HTTPUnauthorized(resp.headers.get("WWW-Authenticate"), url)
         if resp.status_code == 407:
             raise HTTPProxyUnauthorized(resp.headers.get("Proxy-Authenticate"), url)
         if resp.status_code != 200:
-            raise GitProtocolError(
-                "unexpected http resp %d for %s" % (resp.status_code, url)
-            )
+            raise GitProtocolError(f"unexpected http resp {resp.status_code} for {url}")
 
         # Add required fields as stated in AbstractHttpGitClient._http_request
-        resp.content_type = resp.headers.get("Content-Type")
-        resp.redirect_location = ""
+        resp.content_type = resp.headers.get("Content-Type")  # type: ignore[attr-defined]
+        resp.redirect_location = ""  # type: ignore[attr-defined]
         if resp.history:
-            resp.redirect_location = resp.url
+            resp.redirect_location = resp.url  # type: ignore[attr-defined]
 
         read = BytesIO(resp.content).read
 
         return resp, read
 
 
-def get_session(config):
+def get_session(config: Optional["ConfigFile"]) -> Session:
+    """Create a requests session with Git configuration.
+
+    Args:
+      config: Git configuration file
+
+    Returns:
+      Configured requests Session
+    """
     session = Session()
     session.headers.update({"Pragma": "no-cache"})
 
-    proxy_server = user_agent = ca_certs = ssl_verify = None
+    proxy_server: Optional[str] = None
+    user_agent: Optional[str] = None
+    ca_certs: Optional[str] = None
+    ssl_verify: Optional[bool] = None
 
     if config is not None:
         try:
-            proxy_server = config.get(b"http", b"proxy")
-            if isinstance(proxy_server, bytes):
-                proxy_server = proxy_server.decode()
+            proxy_bytes = config.get(b"http", b"proxy")
+            if isinstance(proxy_bytes, bytes):
+                proxy_server = proxy_bytes.decode()
         except KeyError:
             pass
 
         try:
-            user_agent = config.get(b"http", b"useragent")
-            if isinstance(user_agent, bytes):
-                user_agent = user_agent.decode()
+            agent_bytes = config.get(b"http", b"useragent")
+            if isinstance(agent_bytes, bytes):
+                user_agent = agent_bytes.decode()
         except KeyError:
             pass
 
@@ -127,24 +161,22 @@ def get_session(config):
             ssl_verify = True
 
         try:
-            ca_certs = config.get(b"http", b"sslCAInfo")
-            if isinstance(ca_certs, bytes):
-                ca_certs = ca_certs.decode()
+            certs_bytes = config.get(b"http", b"sslCAInfo")
+            if isinstance(certs_bytes, bytes):
+                ca_certs = certs_bytes.decode()
         except KeyError:
             ca_certs = None
 
     if user_agent is None:
         user_agent = default_user_agent_string()
-    session.headers.update({"User-agent": user_agent})
+    if user_agent is not None:
+        session.headers.update({"User-agent": user_agent})
 
     if ca_certs:
         session.verify = ca_certs
     elif ssl_verify is False:
         session.verify = ssl_verify
 
-    if proxy_server:
-        session.proxies.update({
-            "http": proxy_server,
-            "https": proxy_server
-        })
+    if proxy_server is not None:
+        session.proxies.update({"http": proxy_server, "https": proxy_server})
     return session
