@@ -1,8 +1,9 @@
 # __init__.py -- Fast export/import functionality
 # Copyright (C) 2010-2013 Jelmer Vernooij <jelmer@jelmer.uk>
 #
+# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
-# General Public License as public by the Free Software Foundation; version 2.0
+# General Public License as published by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
 # modify it under the terms of either of these two licenses.
 #
@@ -22,16 +23,33 @@
 """Fast export/import functionality."""
 
 import stat
+from collections.abc import Generator
+from typing import TYPE_CHECKING, Any, BinaryIO, Optional
 
 from fastimport import commands, parser, processor
 from fastimport import errors as fastimport_errors
 
 from .index import commit_tree
 from .object_store import iter_tree_contents
-from .objects import ZERO_SHA, Blob, Commit, Tag
+from .objects import ZERO_SHA, Blob, Commit, ObjectID, Tag
+from .refs import Ref
+
+if TYPE_CHECKING:
+    from .object_store import BaseObjectStore
+    from .repo import BaseRepo
 
 
-def split_email(text):
+def split_email(text: bytes) -> tuple[bytes, bytes]:
+    """Split email address from name.
+
+    Args:
+        text: Full name and email (e.g. b"John Doe <john@example.com>")
+
+    Returns:
+        Tuple of (name, email)
+    """
+    # TODO(jelmer): Dedupe this and the same functionality in
+    # format_annotate_line.
     (name, email) = text.rsplit(b" <", 1)
     return (name, email.rstrip(b">"))
 
@@ -39,41 +57,85 @@ def split_email(text):
 class GitFastExporter:
     """Generate a fast-export output stream for Git objects."""
 
-    def __init__(self, outf, store) -> None:
+    def __init__(self, outf: BinaryIO, store: "BaseObjectStore") -> None:
+        """Initialize the fast exporter.
+
+        Args:
+            outf: Output file to write to
+            store: Object store to export from
+        """
         self.outf = outf
         self.store = store
         self.markers: dict[bytes, bytes] = {}
         self._marker_idx = 0
 
-    def print_cmd(self, cmd):
-        self.outf.write(getattr(cmd, "__bytes__", cmd.__repr__)() + b"\n")
+    def print_cmd(self, cmd: object) -> None:
+        """Print a command to the output stream.
 
-    def _allocate_marker(self):
+        Args:
+            cmd: Command object to print
+        """
+        if hasattr(cmd, "__bytes__"):
+            output = cmd.__bytes__()
+        else:
+            output = cmd.__repr__().encode("utf-8")
+        self.outf.write(output + b"\n")
+
+    def _allocate_marker(self) -> bytes:
+        """Allocate a new marker.
+
+        Returns:
+            New marker as bytes
+        """
         self._marker_idx += 1
-        return ("%d" % (self._marker_idx,)).encode("ascii")
+        return str(self._marker_idx).encode("ascii")
 
-    def _export_blob(self, blob):
+    def _export_blob(self, blob: Blob) -> tuple[Any, bytes]:
+        """Export a blob object.
+
+        Args:
+            blob: Blob object to export
+
+        Returns:
+            Tuple of (BlobCommand, marker)
+        """
         marker = self._allocate_marker()
         self.markers[marker] = blob.id
         return (commands.BlobCommand(marker, blob.data), marker)
 
-    def emit_blob(self, blob):
+    def emit_blob(self, blob: Blob) -> bytes:
+        """Emit a blob to the output stream.
+
+        Args:
+            blob: Blob object to emit
+
+        Returns:
+            Marker for the blob
+        """
         (cmd, marker) = self._export_blob(blob)
         self.print_cmd(cmd)
         return marker
 
-    def _iter_files(self, base_tree, new_tree):
+    def _iter_files(
+        self, base_tree: Optional[bytes], new_tree: Optional[bytes]
+    ) -> Generator[Any, None, None]:
         for (
             (old_path, new_path),
             (old_mode, new_mode),
             (old_hexsha, new_hexsha),
         ) in self.store.tree_changes(base_tree, new_tree):
             if new_path is None:
-                yield commands.FileDeleteCommand(old_path)
+                if old_path is not None:
+                    yield commands.FileDeleteCommand(old_path)
                 continue
-            if not stat.S_ISDIR(new_mode):
-                blob = self.store[new_hexsha]
-                marker = self.emit_blob(blob)
+            marker = b""
+            if new_mode is not None and not stat.S_ISDIR(new_mode):
+                if new_hexsha is not None:
+                    blob = self.store[new_hexsha]
+                    from .objects import Blob
+
+                    if isinstance(blob, Blob):
+                        marker = self.emit_blob(blob)
             if old_path != new_path and old_path is not None:
                 yield commands.FileRenameCommand(old_path, new_path)
             if old_mode != new_mode or old_hexsha != new_hexsha:
@@ -82,7 +144,9 @@ class GitFastExporter:
                     new_path, new_mode, prefixed_marker, None
                 )
 
-    def _export_commit(self, commit, ref, base_tree=None):
+    def _export_commit(
+        self, commit: Commit, ref: Ref, base_tree: Optional[ObjectID] = None
+    ) -> tuple[Any, bytes]:
         file_cmds = list(self._iter_files(base_tree, commit.tree))
         marker = self._allocate_marker()
         if commit.parents:
@@ -110,7 +174,19 @@ class GitFastExporter:
         )
         return (cmd, marker)
 
-    def emit_commit(self, commit, ref, base_tree=None):
+    def emit_commit(
+        self, commit: Commit, ref: Ref, base_tree: Optional[ObjectID] = None
+    ) -> bytes:
+        """Emit a commit in fast-export format.
+
+        Args:
+          commit: Commit object to export
+          ref: Reference name for the commit
+          base_tree: Base tree for incremental export
+
+        Returns:
+          Marker for the commit
+        """
         cmd, marker = self._export_commit(commit, ref, base_tree)
         self.print_cmd(cmd)
         return marker
@@ -121,34 +197,64 @@ class GitImportProcessor(processor.ImportProcessor):
 
     # FIXME: Batch creation of objects?
 
-    def __init__(self, repo, params=None, verbose=False, outf=None) -> None:
+    def __init__(
+        self,
+        repo: "BaseRepo",
+        params: Optional[Any] = None,  # noqa: ANN401
+        verbose: bool = False,
+        outf: Optional[BinaryIO] = None,
+    ) -> None:
+        """Initialize GitImportProcessor.
+
+        Args:
+          repo: Repository to import into
+          params: Import parameters
+          verbose: Whether to enable verbose output
+          outf: Output file for verbose messages
+        """
         processor.ImportProcessor.__init__(self, params, verbose)
         self.repo = repo
         self.last_commit = ZERO_SHA
         self.markers: dict[bytes, bytes] = {}
         self._contents: dict[bytes, tuple[int, bytes]] = {}
 
-    def lookup_object(self, objectish):
+    def lookup_object(self, objectish: bytes) -> ObjectID:
+        """Look up an object by reference or marker.
+
+        Args:
+          objectish: Object reference or marker
+
+        Returns:
+          Object ID
+        """
         if objectish.startswith(b":"):
             return self.markers[objectish[1:]]
         return objectish
 
-    def import_stream(self, stream):
+    def import_stream(self, stream: BinaryIO) -> dict[bytes, bytes]:
+        """Import from a fast-import stream.
+
+        Args:
+          stream: Stream to import from
+
+        Returns:
+          Dictionary of markers to object IDs
+        """
         p = parser.ImportParser(stream)
         self.process(p.iter_commands)
         return self.markers
 
-    def blob_handler(self, cmd):
+    def blob_handler(self, cmd: commands.BlobCommand) -> None:
         """Process a BlobCommand."""
         blob = Blob.from_string(cmd.data)
         self.repo.object_store.add_object(blob)
         if cmd.mark:
             self.markers[cmd.mark] = blob.id
 
-    def checkpoint_handler(self, cmd):
+    def checkpoint_handler(self, cmd: commands.CheckpointCommand) -> None:
         """Process a CheckpointCommand."""
 
-    def commit_handler(self, cmd):
+    def commit_handler(self, cmd: commands.CommitCommand) -> None:
         """Process a CommitCommand."""
         commit = Commit()
         if cmd.author is not None:
@@ -177,7 +283,7 @@ class GitImportProcessor(processor.ImportProcessor):
             if filecmd.name == b"filemodify":
                 if filecmd.data is not None:
                     blob = Blob.from_string(filecmd.data)
-                    self.repo.object_store.add(blob)
+                    self.repo.object_store.add_object(blob)
                     blob_id = blob.id
                 else:
                     blob_id = self.lookup_object(filecmd.dataref)
@@ -207,16 +313,21 @@ class GitImportProcessor(processor.ImportProcessor):
         if cmd.mark:
             self.markers[cmd.mark] = commit.id
 
-    def progress_handler(self, cmd):
+    def progress_handler(self, cmd: commands.ProgressCommand) -> None:
         """Process a ProgressCommand."""
 
-    def _reset_base(self, commit_id):
+    def _reset_base(self, commit_id: ObjectID) -> None:
         if self.last_commit == commit_id:
             return
         self._contents = {}
         self.last_commit = commit_id
         if commit_id != ZERO_SHA:
-            tree_id = self.repo[commit_id].tree
+            from .objects import Commit
+
+            commit = self.repo[commit_id]
+            tree_id = commit.tree if isinstance(commit, Commit) else None
+            if tree_id is None:
+                return
             for (
                 path,
                 mode,
@@ -224,7 +335,7 @@ class GitImportProcessor(processor.ImportProcessor):
             ) in iter_tree_contents(self.repo.object_store, tree_id):
                 self._contents[path] = (mode, hexsha)
 
-    def reset_handler(self, cmd):
+    def reset_handler(self, cmd: commands.ResetCommand) -> None:
         """Process a ResetCommand."""
         if cmd.from_ is None:
             from_ = ZERO_SHA
@@ -233,15 +344,15 @@ class GitImportProcessor(processor.ImportProcessor):
         self._reset_base(from_)
         self.repo.refs[cmd.ref] = from_
 
-    def tag_handler(self, cmd):
+    def tag_handler(self, cmd: commands.TagCommand) -> None:
         """Process a TagCommand."""
         tag = Tag()
         tag.tagger = cmd.tagger
         tag.message = cmd.message
-        tag.name = cmd.tag
-        self.repo.add_object(tag)
+        tag.name = cmd.from_
+        self.repo.object_store.add_object(tag)
         self.repo.refs["refs/tags/" + tag.name] = tag.id
 
-    def feature_handler(self, cmd):
+    def feature_handler(self, cmd: commands.FeatureCommand) -> None:
         """Process a FeatureCommand."""
         raise fastimport_errors.UnknownFeature(cmd.feature_name)
