@@ -3,7 +3,7 @@
 #
 # SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
-# General Public License as public by the Free Software Foundation; version 2.0
+# General Public License as published by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
 # modify it under the terms of either of these two licenses.
 #
@@ -27,28 +27,75 @@ on.
 
 import email.parser
 import time
+from collections.abc import Generator
 from difflib import SequenceMatcher
-from typing import BinaryIO, Optional, TextIO, Union
+from typing import (
+    IO,
+    TYPE_CHECKING,
+    BinaryIO,
+    Optional,
+    TextIO,
+    Union,
+)
+
+if TYPE_CHECKING:
+    import email.message
+
+    from .object_store import BaseObjectStore
 
 from .objects import S_ISGITLINK, Blob, Commit
-from .pack import ObjectContainer
 
 FIRST_FEW_BYTES = 8000
 
+DEFAULT_DIFF_ALGORITHM = "myers"
+
+
+class DiffAlgorithmNotAvailable(Exception):
+    """Raised when a requested diff algorithm is not available."""
+
+    def __init__(self, algorithm: str, install_hint: str = "") -> None:
+        """Initialize exception.
+
+        Args:
+            algorithm: Name of the unavailable algorithm
+            install_hint: Optional installation hint
+        """
+        self.algorithm = algorithm
+        self.install_hint = install_hint
+        if install_hint:
+            super().__init__(
+                f"Diff algorithm '{algorithm}' requested but not available. {install_hint}"
+            )
+        else:
+            super().__init__(
+                f"Diff algorithm '{algorithm}' requested but not available."
+            )
+
 
 def write_commit_patch(
-    f, commit, contents, progress, version=None, encoding=None
+    f: IO[bytes],
+    commit: "Commit",
+    contents: Union[str, bytes],
+    progress: tuple[int, int],
+    version: Optional[str] = None,
+    encoding: Optional[str] = None,
 ) -> None:
     """Write a individual file patch.
 
     Args:
+      f: File-like object to write to
       commit: Commit object
-      progress: Tuple with current patch number and total.
+      contents: Contents of the patch
+      progress: tuple with current patch number and total.
+      version: Version string to include in patch header
+      encoding: Encoding to use for the patch
 
     Returns:
       tuple with filename and contents
     """
     encoding = encoding or getattr(f, "encoding", "ascii")
+    if encoding is None:
+        encoding = "ascii"
     if isinstance(contents, str):
         contents = contents.encode(encoding)
     (num, total) = progress
@@ -87,10 +134,12 @@ def write_commit_patch(
 
         f.write(b"Dulwich %d.%d.%d\n" % dulwich_version)
     else:
+        if encoding is None:
+            encoding = "ascii"
         f.write(version.encode(encoding) + b"\n")
 
 
-def get_summary(commit):
+def get_summary(commit: "Commit") -> str:
     """Determine the summary line for use in a filename.
 
     Args:
@@ -98,11 +147,12 @@ def get_summary(commit):
     Returns: Summary string
     """
     decoded = commit.message.decode(errors="replace")
-    return decoded.splitlines()[0].replace(" ", "-")
+    lines = decoded.splitlines()
+    return lines[0].replace(" ", "-") if lines else ""
 
 
 #  Unified Diff
-def _format_range_unified(start, stop) -> str:
+def _format_range_unified(start: int, stop: int) -> str:
     """Convert range to the "ed" format."""
     # Per the diff spec at http://www.unix.org/single_unix_specification/
     beginning = start + 1  # lines start numbering with one
@@ -115,24 +165,23 @@ def _format_range_unified(start, stop) -> str:
 
 
 def unified_diff(
-    a,
-    b,
-    fromfile="",
-    tofile="",
-    fromfiledate="",
-    tofiledate="",
-    n=3,
-    lineterm="\n",
-    tree_encoding="utf-8",
-    output_encoding="utf-8",
-):
-    """difflib.unified_diff that can detect "No newline at end of file" as
-    original "git diff" does.
+    a: list[bytes],
+    b: list[bytes],
+    fromfile: bytes = b"",
+    tofile: bytes = b"",
+    fromfiledate: str = "",
+    tofiledate: str = "",
+    n: int = 3,
+    lineterm: str = "\n",
+    tree_encoding: str = "utf-8",
+    output_encoding: str = "utf-8",
+) -> Generator[bytes, None, None]:
+    """difflib.unified_diff that can detect "No newline at end of file" as original "git diff" does.
 
     Based on the same function in Python2.7 difflib.py
     """
     started = False
-    for group in SequenceMatcher(None, a, b).get_grouped_opcodes(n):
+    for group in SequenceMatcher(a=a, b=b).get_grouped_opcodes(n):
         if not started:
             started = True
             fromdate = f"\t{fromfiledate}" if fromfiledate else ""
@@ -166,7 +215,108 @@ def unified_diff(
                     yield b"+" + line
 
 
-def is_binary(content):
+def _get_sequence_matcher(algorithm: str, a: list[bytes], b: list[bytes]):
+    """Get appropriate sequence matcher for the given algorithm.
+
+    Args:
+        algorithm: Diff algorithm ("myers" or "patience")
+        a: First sequence
+        b: Second sequence
+
+    Returns:
+        Configured sequence matcher instance
+
+    Raises:
+        DiffAlgorithmNotAvailable: If patience requested but not available
+    """
+    if algorithm == "patience":
+        try:
+            from patiencediff import PatienceSequenceMatcher
+
+            return PatienceSequenceMatcher(None, a, b)
+        except ImportError:
+            raise DiffAlgorithmNotAvailable(
+                "patience", "Install with: pip install 'dulwich[patiencediff]'"
+            )
+    else:
+        return SequenceMatcher(a=a, b=b)
+
+
+def unified_diff_with_algorithm(
+    a: list[bytes],
+    b: list[bytes],
+    fromfile: bytes = b"",
+    tofile: bytes = b"",
+    fromfiledate: str = "",
+    tofiledate: str = "",
+    n: int = 3,
+    lineterm: str = "\n",
+    tree_encoding: str = "utf-8",
+    output_encoding: str = "utf-8",
+    algorithm: Optional[str] = None,
+) -> Generator[bytes, None, None]:
+    """Generate unified diff with specified algorithm.
+
+    Args:
+        a: First sequence of lines
+        b: Second sequence of lines
+        fromfile: Name of first file
+        tofile: Name of second file
+        fromfiledate: Date of first file
+        tofiledate: Date of second file
+        n: Number of context lines
+        lineterm: Line terminator
+        tree_encoding: Encoding for tree paths
+        output_encoding: Encoding for output
+        algorithm: Diff algorithm to use ("myers" or "patience")
+
+    Returns:
+        Generator yielding diff lines
+
+    Raises:
+        DiffAlgorithmNotAvailable: If patience algorithm requested but patiencediff not available
+    """
+    if algorithm is None:
+        algorithm = DEFAULT_DIFF_ALGORITHM
+
+    matcher = _get_sequence_matcher(algorithm, a, b)
+
+    started = False
+    for group in matcher.get_grouped_opcodes(n):
+        if not started:
+            started = True
+            fromdate = f"\t{fromfiledate}" if fromfiledate else ""
+            todate = f"\t{tofiledate}" if tofiledate else ""
+            yield f"--- {fromfile.decode(tree_encoding)}{fromdate}{lineterm}".encode(
+                output_encoding
+            )
+            yield f"+++ {tofile.decode(tree_encoding)}{todate}{lineterm}".encode(
+                output_encoding
+            )
+
+        first, last = group[0], group[-1]
+        file1_range = _format_range_unified(first[1], last[2])
+        file2_range = _format_range_unified(first[3], last[4])
+        yield f"@@ -{file1_range} +{file2_range} @@{lineterm}".encode(output_encoding)
+
+        for tag, i1, i2, j1, j2 in group:
+            if tag == "equal":
+                for line in a[i1:i2]:
+                    yield b" " + line
+                continue
+            if tag in ("replace", "delete"):
+                for line in a[i1:i2]:
+                    if not line[-1:] == b"\n":
+                        line += b"\n\\ No newline at end of file\n"
+                    yield b"-" + line
+            if tag in ("replace", "insert"):
+                for line in b[j1:j2]:
+                    if not line[-1:] == b"\n":
+                        line += b"\n\\ No newline at end of file\n"
+                    yield b"+" + line
+
+
+def is_binary(content: bytes) -> bool:
     """See if the first few bytes contain any null characters.
 
     Args:
@@ -175,14 +325,31 @@ def is_binary(content):
     return b"\0" in content[:FIRST_FEW_BYTES]
 
 
-def shortid(hexsha):
+def shortid(hexsha: Optional[bytes]) -> bytes:
+    """Get short object ID.
+
+    Args:
+        hexsha: Full hex SHA or None
+
+    Returns:
+        7-character short ID
+    """
     if hexsha is None:
         return b"0" * 7
     else:
         return hexsha[:7]
 
 
-def patch_filename(p, root):
+def patch_filename(p: Optional[bytes], root: bytes) -> bytes:
+    """Generate patch filename.
+
+    Args:
+        p: Path or None
+        root: Root directory
+
+    Returns:
+        Full patch filename
+    """
     if p is None:
         return b"/dev/null"
     else:
@@ -190,7 +357,12 @@ def patch_filename(p, root):
 
 
 def write_object_diff(
-    f, store: ObjectContainer, old_file, new_file, diff_binary=False
+    f: IO[bytes],
+    store: "BaseObjectStore",
+    old_file: tuple[Optional[bytes], Optional[int], Optional[bytes]],
+    new_file: tuple[Optional[bytes], Optional[int], Optional[bytes]],
+    diff_binary: bool = False,
+    diff_algorithm: Optional[str] = None,
 ) -> None:
     """Write the diff for an object.
 
@@ -201,6 +373,7 @@ def write_object_diff(
       new_file: (path, mode, hexsha) tuple
       diff_binary: Whether to diff files even if they
         are considered binary files by is_binary().
+      diff_algorithm: Algorithm to use for diffing ("myers" or "patience")
 
     Note: the tuple elements should be None for nonexistent files
     """
@@ -209,15 +382,37 @@ def write_object_diff(
     patched_old_path = patch_filename(old_path, b"a")
     patched_new_path = patch_filename(new_path, b"b")
 
-    def content(mode, hexsha):
+    def content(mode: Optional[int], hexsha: Optional[bytes]) -> Blob:
+        """Get blob content for a file.
+
+        Args:
+            mode: File mode
+            hexsha: Object SHA
+
+        Returns:
+            Blob object
+        """
         if hexsha is None:
             return Blob.from_string(b"")
-        elif S_ISGITLINK(mode):
+        elif mode is not None and S_ISGITLINK(mode):
             return Blob.from_string(b"Subproject commit " + hexsha + b"\n")
         else:
-            return store[hexsha]
+            obj = store[hexsha]
+            if isinstance(obj, Blob):
+                return obj
+            else:
+                # Fallback for non-blob objects
+                return Blob.from_string(obj.as_raw_string())
 
-    def lines(content):
+    def lines(content: "Blob") -> list[bytes]:
+        """Split blob content into lines.
+
+        Args:
+            content: Blob content
+
+        Returns:
+            List of lines
+        """
         if not content:
             return []
         else:
@@ -239,17 +434,22 @@ def write_object_diff(
         f.write(binary_diff)
     else:
         f.writelines(
-            unified_diff(
+            unified_diff_with_algorithm(
                 lines(old_content),
                 lines(new_content),
                 patched_old_path,
                 patched_new_path,
+                algorithm=diff_algorithm,
             )
         )
 
 
 # TODO(jelmer): Support writing unicode, rather than bytes.
-def gen_diff_header(paths, modes, shas):
+def gen_diff_header(
+    paths: tuple[Optional[bytes], Optional[bytes]],
+    modes: tuple[Optional[int], Optional[int]],
+    shas: tuple[Optional[bytes], Optional[bytes]],
+) -> Generator[bytes, None, None]:
     """Write a blob diff header.
 
     Args:
@@ -282,13 +482,19 @@ def gen_diff_header(paths, modes, shas):
 
 
 # TODO(jelmer): Support writing unicode, rather than bytes.
-def write_blob_diff(f, old_file, new_file) -> None:
+def write_blob_diff(
+    f: IO[bytes],
+    old_file: tuple[Optional[bytes], Optional[int], Optional["Blob"]],
+    new_file: tuple[Optional[bytes], Optional[int], Optional["Blob"]],
+    diff_algorithm: Optional[str] = None,
+) -> None:
     """Write blob diff.
 
     Args:
       f: File-like object to write to
       old_file: (path, mode, hexsha) tuple (None if nonexisting)
       new_file: (path, mode, hexsha) tuple (None if nonexisting)
+      diff_algorithm: Algorithm to use for diffing ("myers" or "patience")
 
     Note: The use of write_object_diff is recommended over this function.
     """
@@ -297,7 +503,15 @@ def write_blob_diff(f, old_file, new_file) -> None:
     patched_old_path = patch_filename(old_path, b"a")
     patched_new_path = patch_filename(new_path, b"b")
 
-    def lines(blob):
+    def lines(blob: Optional["Blob"]) -> list[bytes]:
+        """Split blob content into lines.
+
+        Args:
+            blob: Blob object or None
+
+        Returns:
+            List of lines
+        """
         if blob is not None:
             return blob.splitlines()
         else:
@@ -313,19 +527,34 @@ def write_blob_diff(f, old_file, new_file) -> None:
     old_contents = lines(old_blob)
     new_contents = lines(new_blob)
     f.writelines(
-        unified_diff(old_contents, new_contents, patched_old_path, patched_new_path)
+        unified_diff_with_algorithm(
+            old_contents,
+            new_contents,
+            patched_old_path,
+            patched_new_path,
+            algorithm=diff_algorithm,
+        )
     )
 
 
-def write_tree_diff(f, store, old_tree, new_tree, diff_binary=False) -> None:
+def write_tree_diff(
+    f: IO[bytes],
+    store: "BaseObjectStore",
+    old_tree: Optional[bytes],
+    new_tree: Optional[bytes],
+    diff_binary: bool = False,
+    diff_algorithm: Optional[str] = None,
+) -> None:
     """Write tree diff.
 
     Args:
       f: File-like object to write to.
+      store: Object store to read from
       old_tree: Old tree id
       new_tree: New tree id
       diff_binary: Whether to diff files even if they
         are considered binary files by is_binary().
+      diff_algorithm: Algorithm to use for diffing ("myers" or "patience")
     """
     changes = store.tree_changes(old_tree, new_tree)
     for (oldpath, newpath), (oldmode, newmode), (oldsha, newsha) in changes:
@@ -335,10 +564,13 @@ def write_tree_diff(f, store, old_tree, new_tree, diff_binary=False) -> None:
             (oldpath, oldmode, oldsha),
             (newpath, newmode, newsha),
             diff_binary=diff_binary,
+            diff_algorithm=diff_algorithm,
         )
 
 
-def git_am_patch_split(f: Union[TextIO, BinaryIO], encoding: Optional[str] = None):
+def git_am_patch_split(
+    f: Union[TextIO, BinaryIO], encoding: Optional[str] = None
+) -> tuple["Commit", bytes, Optional[bytes]]:
     """Parse a git-am-style patch and split it up into bits.
 
     Args:
@@ -358,7 +590,9 @@ def git_am_patch_split(f: Union[TextIO, BinaryIO], encoding: Optional[str] = Non
     return parse_patch_message(msg, encoding)
 
 
-def parse_patch_message(msg, encoding=None):
+def parse_patch_message(
+    msg: "email.message.Message", encoding: Optional[str] = None
+) -> tuple["Commit", bytes, Optional[bytes]]:
     """Extract a Commit object and patch from an e-mail message.
 
     Args:
@@ -367,6 +601,8 @@ def parse_patch_message(msg, encoding=None):
     Returns: Tuple with commit object, diff contents and git version
     """
     c = Commit()
+    if encoding is None:
+        encoding = "ascii"
     c.author = msg["from"].encode(encoding)
     c.committer = msg["from"].encode(encoding)
     try:
@@ -380,7 +616,13 @@ def parse_patch_message(msg, encoding=None):
     first = True
 
     body = msg.get_payload(decode=True)
-    lines = body.splitlines(True)
+    if isinstance(body, str):
+        body = body.encode(encoding)
+    if isinstance(body, bytes):
+        lines = body.splitlines(True)
+    else:
+        # Handle other types by converting to string first
+        lines = str(body).encode(encoding).splitlines(True)
     line_iter = iter(lines)
 
     for line in line_iter:
