@@ -3,7 +3,7 @@
 #
 # SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
-# General Public License as public by the Free Software Foundation; version 2.0
+# General Public License as published by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
 # modify it under the terms of either of these two licenses.
 #
@@ -43,7 +43,7 @@ except ImportError:
 class TagCreateSignTestCase(PorcelainGpgTestCase, CompatTestCase):
     def test_sign(self) -> None:
         # Test that dulwich signatures can be verified by CGit
-        c1, c2, c3 = build_commit_graph(
+        _c1, _c2, c3 = build_commit_graph(
             self.repo.object_store, [[1], [2, 1], [3, 1, 2]]
         )
         self.repo.refs[b"HEAD"] = c3.id
@@ -67,7 +67,7 @@ class TagCreateSignTestCase(PorcelainGpgTestCase, CompatTestCase):
 
     def test_verify(self) -> None:
         # Test that CGit signatures can be verified by dulwich
-        c1, c2, c3 = build_commit_graph(
+        _c1, _c2, c3 = build_commit_graph(
             self.repo.object_store, [[1], [2, 1], [3, 1, 2]]
         )
         self.repo.refs[b"HEAD"] = c3.id
