@@ -3,7 +3,7 @@
 #
 # SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
-# General Public License as public by the Free Software Foundation; version 2.0
+# General Public License as published by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
 # modify it under the terms of either of these two licenses.
 #
@@ -33,7 +33,9 @@ from urllib.parse import urlparse
 
 import dulwich
 from dulwich import client
+from dulwich.bundle import create_bundle_from_repo, write_bundle
 from dulwich.client import (
+    BundleClient,
     FetchPackResult,
     GitProtocolError,
     HangupException,
@@ -59,7 +61,7 @@ from dulwich.client import (
     parse_rsync_url,
 )
 from dulwich.config import ConfigDict
-from dulwich.objects import Commit, Tree
+from dulwich.objects import Blob, Commit, Tree
 from dulwich.pack import pack_objects_to_data, write_pack_data, write_pack_objects
 from dulwich.protocol import DEFAULT_GIT_PROTOCOL_VERSION_FETCH, TCP_GIT_PORT, Protocol
 from dulwich.repo import MemoryRepo, Repo
@@ -422,6 +424,20 @@ class TestGetTransportAndPath(TestCase):
         self.assertEqual(1234, c._port)
         self.assertEqual("/bar/baz", path)
 
+    def test_tcp_ipv6(self) -> None:
+        c, path = get_transport_and_path("git://[::1]/bar/baz")
+        self.assertIsInstance(c, TCPGitClient)
+        self.assertEqual("::1", c._host)
+        self.assertEqual(TCP_GIT_PORT, c._port)
+        self.assertEqual("/bar/baz", path)
+
+    def test_tcp_ipv6_port(self) -> None:
+        c, path = get_transport_and_path("git://[2001:db8::1]:1234/bar/baz")
+        self.assertIsInstance(c, TCPGitClient)
+        self.assertEqual("2001:db8::1", c._host)
+        self.assertEqual(1234, c._port)
+        self.assertEqual("/bar/baz", path)
+
     def test_git_ssh_explicit(self) -> None:
         c, path = get_transport_and_path("git+ssh://foo.com/bar/baz")
         self.assertIsInstance(c, SSHGitClient)
@@ -532,7 +548,7 @@ class TestGetTransportAndPath(TestCase):
         from dulwich.config import ConfigDict
 
         config = ConfigDict()
-        c, path = get_transport_and_path(
+        c, _path = get_transport_and_path(
             "ssh://git@github.com/user/repo.git", config=config
         )
         self.assertIsInstance(c, SSHGitClient)
@@ -540,14 +556,14 @@ class TestGetTransportAndPath(TestCase):
 
         config.set((b"core",), b"sshCommand", b"custom-ssh -o CustomOption=yes")
 
-        c, path = get_transport_and_path(
+        c, _path = get_transport_and_path(
             "ssh://git@github.com/user/repo.git", config=config
         )
         self.assertIsInstance(c, SSHGitClient)
         self.assertEqual("custom-ssh -o CustomOption=yes", c.ssh_command)
 
         # Test rsync-style URL also gets the config
-        c, path = get_transport_and_path("git@github.com:user/repo.git", config=config)
+        c, _path = get_transport_and_path("git@github.com:user/repo.git", config=config)
         self.assertIsInstance(c, SSHGitClient)
         self.assertEqual("custom-ssh -o CustomOption=yes", c.ssh_command)
 
@@ -560,7 +576,7 @@ class TestGetTransportAndPath(TestCase):
     def test_error(self) -> None:
         # Need to use a known urlparse.uses_netloc URL scheme to get the
         # expected parsing of the URL on Python versions less than 2.6.5
-        c, path = get_transport_and_path("prospero://bar/baz")
+        c, _path = get_transport_and_path("prospero://bar/baz")
         self.assertIsInstance(c, SSHGitClient)
 
     def test_http(self) -> None:
@@ -1023,6 +1039,137 @@ class LocalGitClientTests(TestCase):
         self.assertEqual(obj_local, obj_target)
 
 
+class BundleClientTests(TestCase):
+    def setUp(self) -> None:
+        super().setUp()
+        self.tempdir = tempfile.mkdtemp()
+        self.addCleanup(shutil.rmtree, self.tempdir)
+
+    def _create_test_bundle(self):
+        """Create a test bundle file and return its path."""
+        # Create a simple repository
+        repo = MemoryRepo()
+
+        # Create some objects
+        blob = Blob.from_string(b"Hello world")
+        repo.object_store.add_object(blob)
+
+        tree = Tree()
+        tree.add(b"hello.txt", 0o100644, blob.id)
+        repo.object_store.add_object(tree)
+
+        commit = Commit()
+        commit.tree = tree.id
+        commit.message = b"Initial commit"
+        commit.author = commit.committer = b"Test User <test@example.com>"
+        commit.commit_time = commit.author_time = 1234567890
+        commit.commit_timezone = commit.author_timezone = 0
+        repo.object_store.add_object(commit)
+
+        repo.refs[b"refs/heads/master"] = commit.id
+
+        # Create bundle
+        bundle = create_bundle_from_repo(repo)
+
+        # Write bundle to file
+        bundle_path = os.path.join(self.tempdir, "test.bundle")
+        with open(bundle_path, "wb") as f:
+            write_bundle(f, bundle)
+
+        return bundle_path, repo
+
+    def test_is_bundle_file(self) -> None:
+        """Test bundle file detection."""
+        bundle_path, _ = self._create_test_bundle()
+
+        # Test positive case
+        self.assertTrue(BundleClient._is_bundle_file(bundle_path))
+
+        # Test negative case - regular file
+        regular_file = os.path.join(self.tempdir, "regular.txt")
+        with open(regular_file, "w") as f:
+            f.write("not a bundle")
+        self.assertFalse(BundleClient._is_bundle_file(regular_file))
+
+        # Test negative case - non-existent file
+        self.assertFalse(BundleClient._is_bundle_file("/non/existent/file"))
+
+    def test_get_refs(self) -> None:
+        """Test getting refs from bundle."""
+        bundle_path, _ = self._create_test_bundle()
+
+        client = BundleClient()
+        result = client.get_refs(bundle_path)
+
+        self.assertIn(b"refs/heads/master", result.refs)
+        self.assertEqual(result.symrefs, {})
+
+    def test_fetch_pack(self) -> None:
+        """Test fetching pack from bundle."""
+        bundle_path, _source_repo = self._create_test_bundle()
+
+        client = BundleClient()
+        pack_data = BytesIO()
+
+        def determine_wants(refs):
+            return list(refs.values())
+
+        class MockGraphWalker:
+            def next(self):
+                return None
+
+            def ack(self, sha):
+                pass
+
+        result = client.fetch_pack(
+            bundle_path, determine_wants, MockGraphWalker(), pack_data.write
+        )
+
+        # Verify we got refs back
+        self.assertIn(b"refs/heads/master", result.refs)
+
+        # Verify pack data was written
+        self.assertGreater(len(pack_data.getvalue()), 0)
+
+    def test_fetch(self) -> None:
+        """Test fetching from bundle into target repo."""
+        bundle_path, _source_repo = self._create_test_bundle()
+
+        client = BundleClient()
+        target_repo = MemoryRepo()
+
+        result = client.fetch(bundle_path, target_repo)
+
+        # Verify refs were imported
+        self.assertIn(b"refs/heads/master", result.refs)
+
+        # Verify objects were imported
+        master_id = result.refs[b"refs/heads/master"]
+        self.assertIn(master_id, target_repo.object_store)
+
+        # Verify the commit object is correct
+        commit = target_repo.object_store[master_id]
+        self.assertEqual(commit.message, b"Initial commit")
+
+    def test_send_pack_not_supported(self) -> None:
+        """Test that send_pack raises NotImplementedError."""
+        bundle_path, _ = self._create_test_bundle()
+
+        client = BundleClient()
+
+        with self.assertRaises(NotImplementedError):
+            client.send_pack(bundle_path, None, None)
+
+    def test_get_transport_and_path_bundle(self) -> None:
+        """Test that get_transport_and_path detects bundle files."""
+        bundle_path, _ = self._create_test_bundle()
+
+        client, path = get_transport_and_path(bundle_path)
+
+        self.assertIsInstance(client, BundleClient)
+        self.assertEqual(path, bundle_path)
+
+
 class HttpGitClientTests(TestCase):
     def test_get_url(self) -> None:
         base_url = "https://github.com/jelmer/dulwich"
@@ -1456,6 +1603,40 @@ class HttpGitClientTests(TestCase):
         c = HttpGitClient(url, config=config, timeout=15)
         self.assertEqual(c._timeout, 15)
 
+    def test_http_extra_headers_from_config(self) -> None:
+        """Test that http.extraHeader config values are applied."""
+        from dulwich.config import ConfigDict
+
+        url = "https://github.com/jelmer/dulwich"
+        config = ConfigDict()
+        # Set a single extra header
+        config.set((b"http",), b"extraHeader", b"X-Custom-Header: test-value")
+
+        c = HttpGitClient(url, config=config)
+        # Check that the header was added to the pool manager
+        self.assertIn("X-Custom-Header", c.pool_manager.headers)
+        self.assertEqual(c.pool_manager.headers["X-Custom-Header"], "test-value")
+
+    def test_http_multiple_extra_headers_from_config(self) -> None:
+        """Test that multiple http.extraHeader config values are applied."""
+        from dulwich.config import ConfigDict
+
+        url = "https://github.com/jelmer/dulwich"
+        config = ConfigDict()
+        # Set multiple extra headers
+        config.set((b"http",), b"extraHeader", b"X-Header-1: value1")
+        config.add((b"http",), b"extraHeader", b"X-Header-2: value2")
+        config.add((b"http",), b"extraHeader", b"Authorization: Bearer token123")
+
+        c = HttpGitClient(url, config=config)
+        # Check that all headers were added to the pool manager
+        self.assertIn("X-Header-1", c.pool_manager.headers)
+        self.assertEqual(c.pool_manager.headers["X-Header-1"], "value1")
+        self.assertIn("X-Header-2", c.pool_manager.headers)
+        self.assertEqual(c.pool_manager.headers["X-Header-2"], "value2")
+        self.assertIn("Authorization", c.pool_manager.headers)
+        self.assertEqual(c.pool_manager.headers["Authorization"], "Bearer token123")
+
 
 class TCPGitClientTests(TestCase):
     def test_get_url(self) -> None:
@@ -1475,6 +1656,32 @@ class TCPGitClientTests(TestCase):
         url = c.get_url(path)
         self.assertEqual("git://github.com:9090/jelmer/dulwich", url)
 
+    def test_get_url_with_ipv6(self) -> None:
+        host = "::1"
+        path = "/jelmer/dulwich"
+        c = TCPGitClient(host)
+
+        url = c.get_url(path)
+        self.assertEqual("git://[::1]/jelmer/dulwich", url)
+
+    def test_get_url_with_ipv6_and_port(self) -> None:
+        host = "2001:db8::1"
+        path = "/jelmer/dulwich"
+        port = 9090
+        c = TCPGitClient(host, port=port)
+
+        url = c.get_url(path)
+        self.assertEqual("git://[2001:db8::1]:9090/jelmer/dulwich", url)
+
+    def test_get_url_with_ipv6_default_port(self) -> None:
+        host = "2001:db8::1"
+        path = "/jelmer/dulwich"
+        port = TCP_GIT_PORT  # Default port should not be included in URL
+        c = TCPGitClient(host, port=port)
+
+        url = c.get_url(path)
+        self.assertEqual("git://[2001:db8::1]/jelmer/dulwich", url)
+
 
 class DefaultUrllib3ManagerTest(TestCase):
     def test_no_config(self) -> None:
